#!/usr/bin/perl -w
use strict;
use Math::Trig;  # so we can use acos(-1) to define pi

# List of mineral names
my @minerals = (
		"0600_Anhydrite",
		"0600_Barite",
		"0600_Chalcopyrite",
		"0600_Pyrite",
		"0600_Pyrrhotite",
		"0600_Sphalerite",
		"0600_Wurtzite",
		"0100_White",
		"0300_Water",
	       );

# Set up the paths as the base location in the source tree.
my $path = `pwd`;
chomp $path;

my @volfiles = glob("$path/data/*.vol");
die "No volume files\n" if scalar(@volfiles) < 1;

# Make sure that the output directory exists
mkdir "$path/images/";

my %params;
{ local *CONTROL;
  open(CONTROL, "$path/configuration/control.cfg") or 
    die "Can't open config file\n";
  # Parse the config values!
  while (<CONTROL>) {
    chomp;
    s/\s*//g;
    my ($param, $values) = split(/=/);
    $values =~ s/[{};]//g;
    $params{$param} = [ split(/,/, $values) ];
  }
  close CONTROL;
}

# Build an array of available volume files
# Use dataname as a hash key in datalist to access associated data
my %datalist;
for my $volfile (@volfiles) {
  # data info is (full_filename, dim1, dim2, dim3, pos1, pos2, pos3)
  my ($dataname) = ($volfile =~ m|.*/([^\./]*)\.[^\.]*\.[^\.]*.vol|g);
  my @temp = $volfile =~ m/[^\.]*\.([^x]*)x([^x]*)x([^x]*)\.([^_]*)_([^_]*)_([^_]*)\.vol/g;
  $datalist{$dataname} = [ $volfile,  $volfile =~ m|\/$dataname\.([^x]*)x([^x]*)x([^x]*)\.([^_]*)_([^_]*)_([^_]*)\.vol|g ];
}


# Initialize the lights
my @lightnames = qw/ top farleft farright left right /;
my @lightpositions; 
my @lightfoci;

# These files have info on light positions and which lights should be used
# They will automatically be set up by the script
my $lightlist = "autolightlist";



# The outer loops run the entire rendering pipeline
# In general, these variable are needed for both light propagation and
# rendering
for my $d (@{$params{'data'}}) {
  for my $c (@{$params{'conversion'}}) {
    for my $s (@{$params{'scale'}}) {
      for my $rw (@{$params{'rw'}}) {
	
	# This part handles upsampled volumes
	my $enh = 1;
	$enh = $1 if ($d =~ m/_enhx([\d]*)/g);
	$s /= $enh;
	
	# Set up the light positioning info
	my @datainfo = @{$datalist{$d}};
	my $dimStr = $datainfo[1]." ".$datainfo[2]." ".$datainfo[3];
	my $posStr = $datainfo[4]." ".$datainfo[5]." ".$datainfo[6];
	my @lightcenter = ($datainfo[4], $datainfo[5], $datainfo[6]);
	$lightpositions[0] = (1-$lightcenter[0]).' '.$lightcenter[1].' '.($lightcenter[2]+3);
	$lightfoci[0] = $lightcenter[0].' '.(1-$lightcenter[1]).' '.($lightcenter[2]+2.5);

	$lightpositions[1] = ($lightcenter[0]-10).' '.$lightcenter[1].' '.($lightcenter[2]-3);
	$lightfoci[1] = ($lightcenter[0]-8).' '.(1-$lightcenter[1]).' '.($lightcenter[2]-2);

	$lightpositions[2] = ($lightcenter[0]+10).' '.$lightcenter[1].' '.($lightcenter[2]-3);
	$lightfoci[2] = ($lightcenter[0]+8).' '.(1-$lightcenter[1]).' '.($lightcenter[2]-2);

	$lightpositions[3] = ($lightcenter[0]-2).' '.$lightcenter[1].' '.($lightcenter[2]-2);
	$lightfoci[3] = ($lightcenter[0]-1).' '.(1-$lightcenter[1]).' '.($lightcenter[2]-1);

	$lightpositions[4] = ($lightcenter[0]+2).' '.$lightcenter[1].' '.($lightcenter[2]-2);
	$lightfoci[4] = ($lightcenter[0]+1).' '.(1-$lightcenter[1]).' '.($lightcenter[2]-1);

	my $numlights = scalar @{$params{'lights'}}; # used for volRen too.
	# Generate the lightlist file and perform light propagation phase
	print "Generating... ";
	
	# The inner loops run the image rendering
	for my $m (@{$params{'mineral'}}) {
	  for my $am (@{$params{'ambient'}}) {
	    for my $df (@{$params{'diffuse'}}) {
	      for my $rm (@{$params{'rm'}}) {
		for my $cs (@{$params{'colorscale'}}) {
		  for my $bg (@{$params{'bg'}}) {
		    for my $angle (@{$params{'angle'}}) {
			
		      my ($mName) = ($m =~ m/.*_([^_]*)/g);
		      my ($bgName) = ($bg =~ m/.*_([^_]*)/g);
		      
		      my $dist = 20;
		      my $rads = $angle * acos(-1) / 180;
		      my $xdisp = $dist * sin($rads);
		      my $ydisp = $dist * (1 - cos($rads));
		      my $zdisp = 0;
		      
		      # two-phase (tp)
		      # source dataset
		      #    (some data names end in _1, or something similar, 
		      # which may look confusing)
		      # number of wavelengths (at this time, always will be 31 or 3)
		      # number of lights (L)
		      # data conversion factor (c)
		      # scale of voxels in meters (ms)
		      # "mineral" data used for particle reflections (min)  // currently render-only
		      # ambient factor (a)  // currently render-only
		      # diffuse factor (d)  // currently render-only
		      # percentage of light the minerals reflect (rm)  // currently render-only
		      # percentage of light the water reflects (rw)  // currently render-only
		      # scaling factor applied to final accumulated color (cs)  // render-only
		      # bg color (bg)  // render-only
		      # angle of view from light (0 degrees if omitted)
		      
		      # Format the output filename and build the command line text
		      $c = sprintf("%06d",$c);
		      $s = sprintf("%.2f",$s);
		      $rm = sprintf("%.3f",$rm);
		      $rw = sprintf("%.3f",$rw);
		      my ($anglename, $offset);
		      if ($angle > 0) {
			$anglename = sprintf("_%03d",$angle);
			$offset = "-o $xdisp $ydisp $zdisp ";
		      } else {
			$anglename = "";
			$offset = "";
		      }
		      my $name = "tp_".$d."_31_L".$numlights."_c".$c."_ms".$s.
			"_min".$m."_a".$am."_d".$df."_rm".$rm."_rw".$rw."_cs".$cs."_bg$bgName".$anglename;
		      
		      my $dataFile = $datainfo[0];
		      my $imageFile = "$path/images/$name";

### "observer' should be part of the light definition.
		      for my $light (@{$params{'lights'}}) {
			my $observer = $lightfoci[$light];
			my $shadowfile = "$path/tempsrc/$lightnames[$light]";

			my $configuration = "
Dimensions $dimStr
LightCenter ( $posStr )
DataFile $dataFile
Conversion $c  
ReflFactorW $rw
ReflFactorM $rm  
ScaleFactor $s
ambient $am
diffuse $df
enh $enh
colorscale $cs
light ( $lightpositions[$light] ) ( $lightfoci[$light] ) $shadowfile
ImageFile $imageFile.ppm
Image 300 300
exit
";
			local *CFG;
			open(CFG, ">$path/tempsrc/kls.cfg") or 
			  die "Can't open $path/tempsrc/kls.cfg file\n";
			print CFG $configuration;
			close(CFG);
			
			system "cat $path/tempsrc/kls.cfg | $path/shadowVol";
			
			# Render the image!
			print "Rendering... \n";
			system " cat $path/tempsrc/kls.cfg | $path/volRen ";
			
			# Convert from the default ppm file output to a png
			print "Converting...\n";
			$ENV{PATH} = $ENV{PATH} . ":/sw/bin/";
			system "pnmtopng $imageFile.ppm > $imageFile.png" and 
			  die "Failed to convert ppm to png\n  Is pnmtopng in your path?";
			unlink "$imageFile.ppm" or die "Failed to delete the ppm\n";
		      } # loop over lights
		    }
		  }
		}
	      }
	    }
	  }
	}
      } 
    }
  }
} # outer loops (generate shadow volumes)

