/* ==================================================================== */
/* File : costal_rgb.h --- sub-header file for the "COlour Science and	*/
/* 	Technology Application Library", which provides a set of	*/
/*	commonly used RGB to XYZ conversion matrices. Useful sources	*/
/*	include Adrian Ford and Alan Roberts "Colour Space Conversions"	*/
/*	and Charles Poynton "Frequently Asked Questions about Color",	*/
/*	both of which can be found on the Internet.			*/
/* Author: M. Chen							*/
/* Dates: 10 July 2004							*/
/* ==================================================================== */

/* --------------------------------------- */
/* Default XYZ and RGB Conversion Matrices */
/* --------------------------------------- */

#define M_XYZ2RGB	M_XYZ2RGB_CRT
#define M_RGB2XYZ	M_RGB2XYZ_CRT

/* -------------------------------------------------------------------- */
/* Glassner's matrices, which is based on the NTSC standard for		*/
/* for monitors, with the white point is Illuminant D65.		*/
/* The chromaticity co-ordinates are:					*/
/* R (x=0.67, y=0.33), G (x=0.21, y=0.71), B (x=0.14, y=0.08),		*/
/* White (x=0.312713, y=0.329016.					*/
/* Source: Glassner "How to derive a spectrum from an RGB triplet" and	*/
/* Hall "Illumination and Color in Computer Generated Imagery"		*/
/* -------------------------------------------------------------------- */

// XYZ to sRGB matrix given at
// http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
const double M_XYZ2SRGB[3][3] =
{{ 3.24071,    -0.969258,    0.0556352 },
{ -1.53726,     1.87599,    -0.203996 },
{ -0.498571,    0.0415557,   1.05707 } };

// XYZ to RGB matrix defined in Glassner's book
const double M_XYZ2RGB_CRT[3][3] =
      {  { 1.967, -0.955,  0.064 },
	{ -0.548,  1.938, -0.130 },
	{ -0.297, -0.027,  0.982 } };

// RGB to XYZ matrix obtained by inverting M_XYZ2RGB
const double M_RGB2XYZ_CRT[3][3] =
      { { 0.58929610, 0.29039152, 0.00003661 },
	{ 0.17891824, 0.60511609, 0.06844636 },
	{ 0.18314840, 0.10446478, 1.02022294 } };

/* -------------------------------------------------------------------- */
/* European Broadcasting Union (EBU) uses a Y'U'V' components,		*/
/* with which white point is D65. The chromaticity co-ordinates are:	*/
/* R (x=0.64, y=0.33), G (x=0.29, y=0.60), B (x=0.15, y=0.06),		*/
/* White (x=0.312713, y=0.329016.					*/
/* Source: Adrian Ford and Alan Roberts, "Colour Space Conversions".	*/
/* -------------------------------------------------------------------- */

// XYZ to RGB(EBU) defined with European Y'U'V'.
const double M_XYZ2RGB_EBU[3][3] =
      {  { 3.063, -1.393, -0.476 },
	{ -0.969,  1.876,  0.042 },
	 { 0.068, -0.229,  1.069 } };

// RGB(EBU) to XYZ defined with European Y'U'V'.
const double M_RGB2XYZ_EBU[3][3] =
      { { 0.431, 0.342, 0.178 },
	{ 0.222, 0.707, 0.071 },
	{ 0.020, 0.130, 0.939 } };

/* -------------------------------------------------------------------- */
/* American TV (NTSC coded) uses a Y'I'Q' components, which is no	*/
/* longer in wide use. The CRT gamma law is assumed to be 2.2. The	*/
/* white point is Illuminant C.	The chromaticity co-ordinates are:	*/
/* R (x=0.67, y=0.33), G (x=0.21, y=0.71), B (x=0.14, y=0.08),		*/
/* White (x=0.310063, y=0.316158.					*/
/* Source: Adrian Ford and Alan Roberts, "Colour Space Conversions".	*/
/* -------------------------------------------------------------------- */

// XYZ to RGB(NTSC) defined with American Y'I'Q'.
const double M_XYZ2RGB_NTSC[3][3] =
      {  { 1.910, -0.532, -0.288 },
	{ -0.985,  1.999, -0.028 },
	 { 0.058, -0.118,  0.898 } };

// RGB(NTSC) to XYZ defined with American Y'I'Q'.
const double M_RGB2XYZ_NTSC[3][3] =
      { { 0.607, 0.174, 0.200 },
	{ 0.299, 0.587, 0.114 },
	{ 0.000, 0.066, 1.116 } };

/* -------------------------------------------------------------------- */
/* American TV (SMPTE-C standard) uses a Y'I'Q' components, replacing 	*/
/* the old NTSC standard. The CRT gamma law is assumed to be 2.2. The	*/
/* white point is D65.	The chromaticity co-ordinates are:		*/
/* R (x=0.630, y=0.340), G (x=0.310, y=0.595), B (x=0.155, y=0.070),	*/
/* White (x=0.312713, y=0.329016.					*/
/* Source: Adrian Ford and Alan Roberts, "Colour Space Conversions".	*/
/* -------------------------------------------------------------------- */

// XYZ to RGB(SMPTE) defined with American Y'I'Q'.
const double M_XYZ2RGB_SMPTE[3][3] =
      {  { 3.5058, -1.7397, -0.5440 },
	{ -1.0690,  1.9778, -0.0352 },
	 { 0.0563, -0.1970,  1.0501 } };

// RGB(SMPTE) to XYZ defined with American Y'I'Q'.
const double M_RGB2XYZ_SMPTE[3][3] =
      { { 0.3935, 0.3653, 0.1916 },
	{ 0.2124, 0.7011, 0.0866 },
	{ 0.0187, 0.1119, 0.9582 } };

/* -------------------------------------------------------------------- */
/* ITU has an interim standard for HDTV that uses a Y'CbCr components.	*/
/* The CRT gamma law is assumed to be 2.2. The white point is D65.	*/
/* The standard is ITU.BT-709, and the RGB colour system is sometimes	*/
/* referred to as RGB709 and sRGB. The chromaticity co-ordinates are:	*/
/* R (x=0.64, y=0.33), G (x=0.30, y=0.60), B (x=0.15, y=0.06),		*/
/* White (x=0.312713, y=0.329016.					*/
/* Source: Adrian Ford and Alan Roberts, "Colour Space Conversions" and	*/
/* Charles Poynton "Frequently Asked Questions about Color".		*/
/* -------------------------------------------------------------------- */

// XYZ to RGB(709) defined with ITU.BT-709.
const double M_XYZ2RGB_709[3][3] =
      {  { 3.240479, -1.537150, -0.498535 },
	{ -0.969256,  1.875992, -0.041556 },
	 { 0.055648, -0.204043,  1.057311 } };

// RGB(709) to XYZ defined with ITU-BT-709.
const double M_RGB2XYZ_709[3][3] =
      { { 0.412453, 0.357580, 0.180423 },
	{ 0.212671, 0.715160, 0.072169 },
	{ 0.019334, 0.119193, 0.950227 } };

/* -------------------------------------------------------------------- */
/* SMPTE has a standard for NTSC coding that uses a Y'PbPr components.	*/
/* The CRT gamma law is assumed to be 2.2. The white point is D65.	*/
/* The RGB colour system is sometimes referred to as RGB240. The	*/
/* chromaticity co-ordinates are:					*/
/* R (x=0.67, y=0.33), G (x=0.21, y=0.71), B (x=0.15, y=0.06),		*/
/* White (x=0.312713, y=0.329016.					*/
/* Source: Adrian Ford and Alan Roberts, "Colour Space Conversions".	*/
/* -------------------------------------------------------------------- */

// XYZ to RGB(240) defined with American Y'I'Q'.
const double M_XYZ2RGB_240[3][3] =
      {  { 2.042, -0.565, -0.345 },
	{ -0.894,  1.815, -0.032 },
	 { 0.064, -0.129,  0.912 } };

// RGB(240) to XYZ defined with American Y'I'Q'.
const double M_RGB2XYZ_240[3][3] =
      { { 0.567, 0.190, 0.193 },
	{ 0.279, 0.643, 0.077 },
	{ 0.000, 0.073, 1.016 } };
