/* ============================================================ */
/* File : costal_spd.c --- a set of common subroutines for 	*/
/*	analysis a spectral distribution.			*/
/* Author: M. Chen						*/
/* Dates: 20 June 2004						*/
/* ============================================================ */

#include "costal.h"

/* --------------------------------------- */
/* set SPD31 to a flat grey scale spectrum */
/* --------------------------------------- */

SPD31_set(SPD, v)
double	SPD[31], v;
{
	int	i;

	for (i=0; i<31; i++)
		SPD[i] = v;
}

/* ------------- */
/* copy an SPD31 */
/* ------------- */

SPD31_copy(SPDa, SPDb)
double	SPDa[31], SPDb[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDb[i] = SPDa[i];
}

/* ------------------------------------- */
/* subtract a white spectrum by an SPD31 */
/* ------------------------------------- */

SPD31_inverse(SPDa, SPDb)
double	SPDa[31], SPDb[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDb[i] = 1.0 - SPDa[i];
}

/* ------------------ */
/* scale a SPD31 by s */
/* ------------------ */

SPD31_scale(SPDa, s, SPDb)
double	SPDa[31], s, SPDb[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDb[i] = SPDa[i] * s;
}

/* ------------- */
/* add two SPD31 */
/* ------------- */

SPD31_add(SPDa, SPDb, SPDc)
double	SPDa[31], SPDb[31], SPDc[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDc[i] = SPDa[i] + SPDb[i];
}

/* ------------------------------- */
/* subtract one SPD31 from another */
/* ------------------------------- */

SPD31_sub(SPDa, SPDb, SPDc)
double	SPDa[31], SPDb[31], SPDc[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDc[i] = SPDa[i] - SPDb[i];
}

/* ------------------------------------- */
/* multiply two SPD31 element by element */
/* ------------------------------------- */

SPD31_times(SPDa, SPDb, SPDc)
double	SPDa[31], SPDb[31], SPDc[31];
{
	int	i;

	for (i=0; i<31; i++)
		SPDc[i] = SPDa[i] * SPDb[i];
}

/* --------------------------------------------- */
/* divid one SPD31 by another element by element */
/* --------------------------------------------- */

SPD31_div(SPDa, SPDb, SPDc)
double	SPDa[31], SPDb[31], SPDc[31];
{
	int	i;

	for (i=0; i<31; i++)
		if (IsZero(SPDb[i]))
			SPDc[i] = 1.0e+10;
		else	SPDc[i] = SPDa[i] / SPDb[i];
}

/* ------------------------------------------------ */
/* Counting negative or positive energy in an SPD31 */ 
/* ------------------------------------------------ */

SPD31_negative_energy(SPD, count, sum)
double	SPD[31];
int	*count;
double	*sum;
{
	int	i, _count;
	double	_sum;

	_count = 0;	_sum = 0.0;
	for (i=0; i<31; i++)
		if (SPD[i] < 0) {
			_count++;
			_sum += (-SPD[i]);
		}
	*count = _count;	*sum = _sum;
}

SPD31_positive_energy(SPD, count, sum)
double	SPD[31];
int	*count;
double	*sum;
{
	int	i, _count;
	double	_sum;

	_count = 0;	_sum = 0.0;
	for (i=0; i<31; i++)
		if (SPD[i] > 0) {
			_count++;
			_sum += SPD[i];
		}
	*count = _count;	*sum = _sum;
}

/* ---------------------------------- */
/* Reduce negative energy in an SPD31 */ 
/* ---------------------------------- */

SPD31_improve(SPDa, SPDb)
double	SPDa[31], SPDb[31];
{
	int	i, count1, count2;
	double	X, Y, Z, sum1, sum2;
	double	SPDnp[31], SPDnn[31], SPDnew[31];

	SPD31_negative_energy(SPDa, &count1, &sum1);
	if (count1 <= 0)		// no negative energy
		return(0);

	for (i=0; i<31; i++)
		if (SPDa[i] < 0) {
			SPDnp[i] = -SPDa[i];
			SPDnn[i] = SPDa[i];
		} else
			SPDnp[i] = SPDnn[i] = 0.0;

	SPD31_to_XYZ(SPDnp, &X, &Y, &Z);
	G2_XYZ_to_SPD31(-X, -Y, -Z, SPDnew);
	SPD31_negative_energy(SPDnew, &count2, &sum2);
	if (sum2 < sum1) {
		for (i=0; i<31; i++)
			SPDb[i] = SPDa[i] + SPDnp[i] + SPDnew[i];
		return(1);
	}
/*
	SPD31_to_XYZ(SPDnn, &X, &Y, &Z);
	G2_XYZ_to_SPD31(X, Y, Z, SPDnew);
	SPD31_negative_energy(SPDnew, &count2, &sum2);
	if (sum2 < sum1) {
		for (i=0; i<31; i++)
			SPDb[i] = SPDa[i] - SPDnn[i] + SPDnew[i];
		return(-1);
	}
*/

	return(0);	// cannot improve
}

/* ------------------------------------- */
/* Energy Analysis of SPD31 (400-700:10) */
/* ------------------------------------- */

SPD31_energy(SPD, min, max, avg, pos, neg)
double	SPD[31];
double	*min, *max, *avg, *pos, *neg;
{
	int	i;
	double	_min, _max, _pos, _neg;

	_min = MAXVALUE;
	_max = -MAXVALUE;
	_pos = _neg = 0.0;

	for (i=0; i<31; i++) {
		if (SPD[i] < _min)	_min = SPD[i];
		if (SPD[i] > _max)	_max = SPD[i];
		if (SPD[i] < 0)		_neg += SPD[i];
		if (SPD[i] > 0)		_pos += SPD[i];
	}

	*min = _min;	*max = _max;
	*pos = _pos;	*neg = _neg;
	*avg = (_pos + _neg) / (double)31.0;
}

/* ------------------------------------ */
/* Print an SPD with its energy summary */
/* ------------------------------------ */

SPD31_print(SPD)
double	SPD[31];
{
	int	i;
	double	min, max, avg, pos, neg;

	SPD31_energy(SPD, &min, &max, &avg, &pos, &neg);

	printf("------------------------------------------------\n");
	printf("(%d)%8.4f,  ", 400, SPD[0]);
	printf("(%d)%8.4f,  ", 480, SPD[8]);
	printf("(%d)%8.4f,  ", 560, SPD[16]);
	printf("(%d)%8.4f,  ", 640, SPD[24]);
	printf("\n");

	printf("(%d)%8.4f,  ", 410, SPD[1]);
	printf("(%d)%8.4f,  ", 490, SPD[9]);
	printf("(%d)%8.4f,  ", 570, SPD[17]);
	printf("(%d)%8.4f,  ", 650, SPD[25]);
	printf("\n");

	printf("(%d)%8.4f,  ", 420, SPD[2]);
	printf("(%d)%8.4f,  ", 500, SPD[10]);
	printf("(%d)%8.4f,  ", 580, SPD[18]);
	printf("(%d)%8.4f,  ", 660, SPD[26]);
	printf("\n");

	printf("(%d)%8.4f,  ", 430, SPD[3]);
	printf("(%d)%8.4f,  ", 510, SPD[11]);
	printf("(%d)%8.4f,  ", 590, SPD[19]);
	printf("(%d)%8.4f,  ", 670, SPD[27]);
	printf("\n");

	printf("(%d)%8.4f,  ", 440, SPD[4]);
	printf("(%d)%8.4f,  ", 520, SPD[12]);
	printf("(%d)%8.4f,  ", 600, SPD[20]);
	printf("(%d)%8.4f,  ", 680, SPD[28]);
	printf("\n");

	printf("(%d)%8.4f,  ", 450, SPD[5]);
	printf("(%d)%8.4f,  ", 530, SPD[13]);
	printf("(%d)%8.4f,  ", 610, SPD[21]);
	printf("(%d)%8.4f,  ", 690, SPD[29]);
	printf("\n");

	printf("(%d)%8.4f,  ", 460, SPD[6]);
	printf("(%d)%8.4f,  ", 540, SPD[14]);
	printf("(%d)%8.4f,  ", 620, SPD[22]);
	printf("(%d)%8.4f,  ", 700, SPD[30]);
	printf("\n");

	printf("(%d)%8.4f,  ", 470, SPD[7]);
	printf("(%d)%8.4f,  ", 550, SPD[15]);
	printf("(%d)%8.4f,  ", 630, SPD[23]);
	printf("\n");

//	for (i=0; i<31; i++) {
//		printf("(%d)%8.4f,  ", i*10+400, SPD[i]);
//		if ((i+1)%4 == 0)	printf("\n");
//	}
	printf("\n");
	printf("------------------------------------------------\n");
	printf("min=%+.4f, max=%+.4f, avg=%+.4f\n", min, max, avg);
	printf("pos=%+.4f, neg=%+.4f, pos:neg=%+.9f\n", pos, neg, -pos/neg);
	printf("------------------------------------------------\n");
}
