/* rtInterp.cpp: get interpolated data values */
#include "rtInterp.h"
#include <cstdlib>


enum interpDirs {DO_X=1, DO_Z=2, DO_Y=4};

static long volx=64, voly=64, volz=64, volxvoly=64*64;
static long relIndex[8]={0};
static bool init=false;

void setInterpDims(const int* dims) {
  volx = dims[0];
  voly = dims[1];
  volz = dims[2];
  volxvoly = volx*voly;
  cout << "relIndex = {0";
  for (int i=1; i<8; i++) {
    if (i & DO_X)
      relIndex[i] += 1;
    if (i & DO_Y)
      relIndex[i] += volx;
    if (i & DO_Z)
      relIndex[i] += volxvoly;
    cout << "," << relIndex[i];
  }
  cout << "}\n";
  cout << "Set interpolation dimensions to (" << volx << "," << voly << "," << volz << ")\n";
  init=true;
}

float getValues(const float* volume, const vector<rtColor>& shadow_volume,/* int* dims,*/
	            	rtPoint sample_point, rtColor *color) {

#if DEBUG > 5
  cout << "I'm in getValues(float*,vector<rtColor>*,int*,rtPoint,rtColor*)"
       << endl;
#endif
  if (!init) {
    cerr << "Can't interpolate without initializing!\n";
    exit(-7);
  }

  int xint=0,yint=0,zint=0;
  long index;
  float xf, yf, zf, xfrac, yfrac, zfrac;
  double corners[8]={0.0}, inter[4];
  const rtColor* colorcorners[8] = {NULL};
  static rtColor colorinter[4];
  char dirs = 0, ndirs=0;

  xf = sample_point.getX();
  yf = sample_point.getY();
  zf = sample_point.getZ();

  xint = static_cast<int>(xf);
  yint = static_cast<int>(yf);
  zint = static_cast<int>(zf);

  xfrac = xf - xint;
  yfrac = yf - yint;
  zfrac = zf - zint;

  index = (zint*voly + yint)*volx + xint;

  if (xint < volx-1) {
    dirs |= DO_X;
    ndirs++;
  }
  if (yint < voly-1) {
    dirs |= DO_Y;
    ndirs++;
  }
  if (zint < volz-1) {
    dirs |= DO_Z;
    ndirs++;
  }

  // These lines will short circuit this function
  color->Set(*(shadow_volume.data() + index));
//   return volume[index];

  corners[0] = volume[index];
  colorcorners[0] = shadow_volume.data() + index;

  if (ndirs > 2) { // most common case by far
    for (int i=1; i<8; i++) {
      corners[i] = volume[index + relIndex[i]];
      colorcorners[i] = colorcorners[0] + relIndex[i];
    }
    for (int i=0, j=0; i<4; i++, j+=2) {
      inter[i] = corners[j] + (corners[j+1] - corners[j])*xfrac;
      colorinter[i].Set(colorcorners[j]->getValues());
      colorinter[i].setInterpolate(*(colorcorners[j+1]), xfrac);
    }
    for (int i=0; i<2; i++) {
      inter[i] = inter[i] + (inter[i+2] - inter[i])*yfrac;
      colorinter[i].setInterpolate(colorinter[i+2], yfrac);
    }
    inter[0] = inter[0] + (inter[1] - inter[0])*zfrac;
    colorinter[0].setInterpolate(colorinter[1], zfrac);

    color->Set(colorinter[0]);
    return inter[0];
  }
  // else

  for (int i=1; i<=dirs; i++) {
    if ((i & dirs) == i) {
      corners[i] = volume[index + relIndex[i]];
      colorcorners[i] = colorcorners[0] + relIndex[i];
    }
  }

  if (ndirs > 1) {
    int addInd, incr;
    float frac1, frac2;
    if (dirs & DO_X) {
      addInd = DO_X;
      frac1 = xfrac;
      if (dirs & DO_Y) {
        incr = DO_Y;
        frac2 = yfrac;
      }
      else {
        incr = DO_Z;
        frac2 = zfrac;
      }
    }
    else {
      addInd = DO_Y;
      incr = DO_Z;
      frac1 = yfrac;
      frac2 = zfrac;
    }
    for (int i=0, j=0; i<2; i++, j+=incr) {
      inter[i] = corners[j] + (corners[j+addInd] - corners[j])*frac1;
      colorinter[i].Set(colorcorners[j]->getValues());
      colorinter[i].setInterpolate(*(colorcorners[j+addInd]), frac1);
    }
    inter[0] = inter[0] + (inter[1] - inter[0])*frac2;
    colorinter[0].setInterpolate(colorinter[1], frac2);

    color->Set(colorinter[0]);
    return inter[0];
  }
  else if (ndirs > 0) {
    int addInd;
    float frac;
    if (dirs & DO_X) {
      addInd = DO_X;
      frac = xfrac;
    }
    else if (dirs & DO_Y) {
      addInd = DO_Y;
      frac = yfrac;
    }
    else {
      addInd = DO_Z;
      frac = zfrac;
    }
    inter[0] = corners[0] + (corners[addInd] - corners[0])*frac;
    color->Set(colorcorners[0]->getValues());
    color->setInterpolate(*(colorcorners[addInd]), frac);

    return inter[0];
  }

  // else
  color->Set(colorcorners[0]->getValues());
  return corners[0];

}

rtVector getNormal(const float* buffer, long index) {
  if (!init) {
    cerr << "Can't get normal without initializing!\n";
    exit(-7);
  }

  const int dirs[3] = {DO_X, DO_Y, DO_Z};
  int dims[3] = {volx, voly, volz};
  int pos[3]={0};
  float diffs[3] = {0}, vals[2];

  pos[0] = index % volx;
  pos[1] = (index / volx) % voly;
  pos[2] = index / volxvoly;

  for (int i=0; i<3; i++) {
    vals[0] = (pos[i] > 0) ? buffer[index - relIndex[dirs[i]]] 
                           : 0.0;
    vals[1] = (pos[i] < dims[i]-1) ? buffer[index + relIndex[dirs[i]]] 
                                   : 0.0;
    diffs[i] = vals[1] - vals[0];
  }

  return rtVector(diffs[0], diffs[1], diffs[2]);

}
