#include "rtCommon/rtPoint.h"
#include "rtCommon/rtColor.h"
#include "rtCommon/rtVector.h"
#include "rtCommon/rtLightSpec.h"
#include "rtCommon/rtInterp.h"
#include "utilities/ddebug.h"
#include "utilities/readFile.h"
#include "svConfig.h"
#include "params.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>


int prepareFrontFace(const svConfig&, const rtPoint&,
		     const rtVector&, 
		     vector<rtColor>&, const float* );
void setDisplacements();
void propagateLight(const svConfig&, int, const rtPoint&, 
		    const float*, vector<rtColor>&);
void writeVolume(const string&, unsigned int, const vector<rtColor>&);

const float angle = 75; // Spotlight's angle width
const double PI = acos(-1.0);

rtColor lightcolor(31);

int incr[3];
float viewmag[3];
char viewsign[3];
int direction;
int planex, planey;
int curr_plane;
vector<rtVector> displacements;


#ifdef DEBUG
static DDebug dlog(true, cerr);
#else
static DDebug dlog(false);
#endif


int main(int, char**) {
  bool exitFlag(false);
  while (!exitFlag) {
    while (!cin) { sleep(60); } 
    svConfig config(cin);
    if (config.error()) {
      dlog << "Error in stdin configuration " << config.error() << endl;
    } else {
      dlog << "Configuration info" << endl << config << endl;

      const unsigned int volume 
	= config.volx() * config.voly() * config.volz();
      
      dlog << "Allocating space for input..... " << endl;
      float* involume = 0;
      readFile(config.dataFile(), volume, involume);
      dlog << "Allocated " << volume * sizeof(float) << " bytes." << endl;
      
      dlog << endl << "Allocating space for output..... " << endl;
      vector<rtColor> outvolume;
      outvolume.clear();
      if (volume > outvolume.max_size()) {
	dlog << "Could not allocate memory for the output volume" << endl;
	exit (1);
      }
      rtColor init(colorLength,false); init.setValues(0.0);
      outvolume.assign(volume,init);
      dlog << "Allocated " << volume << " rtColor's successfully" << endl;
      
      const int dims[3] = { config.volx(), config.voly(), config.volz() };
      setInterpDims(dims);
      
      // Loop over all of the lights in the configuratino file.
      const vector<rtLightSpec> lights = config.lights();
      vector<rtLightSpec>::const_iterator i;
      for (i = lights.begin(); i < lights.end(); i++) {
	const rtPoint lightpos = i->position();
	const string& shadowfile = i->file();
	const rtPoint lookpos = i->lookPt();
	rtVector view(lightpos, lookpos);
	view.Normalize();
	dlog << "Light position: " << lightpos <<  endl;
	dlog << "Lookat position: " << lookpos << endl;
	dlog << "Light direction: " << view << endl;
	dlog << endl;
	
	for (int channel=0; channel<31; channel++) {
	  lightcolor.setChannel(channel, light[channel]/* *colorscale*/);
	}
	if (colorLength==3)
	  lightcolor.Set(lightcolor.getRGB());
	dlog << "Light color: " << lightcolor << endl;
	
	incr[0]=1; 
	incr[1] = config.volx(); 
	incr[2] = config.volx() * config.voly();
	viewmag[0] = fabs(view.getI());
	viewmag[1] = fabs(view.getJ());
	viewmag[2] = fabs(view.getK());
	viewsign[0] = (view.getI() > 0) ? 1 : -1;
	viewsign[1] = (view.getJ() > 0) ? 1 : -1;
	viewsign[2] = (view.getK() > 0) ? 1 : -1;
	if (viewmag[0] > viewmag[1]) {
	  if (viewmag[0] > viewmag[2]) {direction = 0;}
	  else {direction = 2;}
	} else {
	  if (viewmag[1] > viewmag[2]) {direction = 1;}
	  else {direction = 2;}
	}
	
	planex = (direction + 1) % 3;
	planey = (direction + 2) % 3;
	
	int ret = prepareFrontFace(config, lightpos, view,
				   outvolume, involume);
	setDisplacements();
	propagateLight(config, ret, lightpos, involume, outvolume );
	writeVolume(shadowfile, volume,  outvolume);
	
      } // loop over processing a single light source
 
      if (involume) delete [] involume;
      exitFlag = config.exit();
      // specifically on cout for chaining
      cout << config << endl << (config.exit() ? "exit" : "done") << endl;

    } // loop processing a single config instance
  } // while (1) loop, only exit if config.exit() is true.
 
  return 0;
}



int prepareFrontFace(const svConfig& config, const rtPoint& light,
		     const rtVector& view, 
		     vector<rtColor>& outvol, const float* invol){
  const float x = config.volx();
  const float y = config.voly();
  const float z = config.volz();
  const int dims[3] = { x, y, z };
  const float conversion = config.conversion();
  const float reflFactorW = config.reflFactorW();
  const float scalefactor = config.scaleFactor();
  dlog << "I'm in prepareFrontFace(rtVector)" << endl;

  int i,j, lightindex, index;
  int indPos[3];
  rtVector currtolight;
  float distance, currvalue, transparency;
  rtColor absfactor(colorLength, false);

  // Cosine of the spotlight's angle width
  double coslimit = -1*cos(angle*PI/180);
  dlog << "coslimit = " << coslimit << endl;

  // Get the index position of the light, and give it the light's color
  indPos[0] = static_cast<int>(light.getX()) + x/2;
  indPos[1] = static_cast<int>(light.getY()) + y/2;
  indPos[2] = static_cast<int>(light.getZ()) + z/2;
  dlog << "indPos: " << rtPoint(indPos[0],indPos[1],indPos[2]) << endl;
  lightindex = (indPos[2] * y + indPos[1]) * x + indPos[0];
  dlog << "lightindex = " << lightindex << endl
       << "lightcolor = " << lightcolor << endl;
  outvol[lightindex].Set(lightcolor);

  if (fabs(light.getX()) > x/2 || 
      fabs(light.getY()) > y/2 ||
      fabs(light.getZ()) > z/2) {
    dlog << "The light is outside the data!\t" << light << endl;
    exit(1);
  }

  int point[3];

  // Set up the first plane in front of the light
  curr_plane = indPos[direction]+viewsign[direction];
  index = curr_plane*incr[direction];
  // For each voxel in the plane
  for (i=0; i<dims[planex]; i++) {
    for (j=0; j<dims[planey]; j++) {
      point[direction] = curr_plane-dims[direction]/2;
      point[planex] = i-dims[planex]/2;
      point[planey] = j-dims[planey]/2;

      // Get the light direction and distance to the light
      currtolight.Set(rtPoint(point[0], point[1], point[2]),
                      light);
      distance = currtolight.getMagnitude();
      currtolight.Normalize();

        // Do nothing if outside the spotlight area
      if (view.dot(currtolight) < coslimit) {
          // Get the local data value
        currvalue = invol[index];
          // precalculate the scaled distance
        float dist = scalefactor*distance;
          // (1-waterReflTrans) is the amount of reflection by the water
        float waterReflTrans = exp(-1*dist*reflFactorW);
          // "transparency" is the proportion of light that doesn't hit the
          // particles (left term) or the water (right term)
        transparency = exp(-1*currvalue*conversion*dist)*waterReflTrans;
          // absfactor is the proportion of light that isn't absorbed by the water
        for (unsigned int channel=0; channel<colorLength; channel++) {
          absfactor.setChannel(channel,
                               exp(-1*absp[0][colorInd[channel]]*dist));
        }
        outvol[index] += lightcolor; // more code, but operations are done in place, without creating and destroying "rtColor"s
        outvol[index] *= absfactor;
        outvol[index] *= transparency;
      }
      index += incr[planey];
    }
    index += incr[planex] - dims[planey]*incr[planey];
  }

  curr_plane += viewsign[direction];
  dlog << "prepareFrontFace(rtVector) completed" << endl;
  return direction;

}

void setDisplacements() {
  displacements.reserve(8);

    // Why 0.3?
  float d=0.3;
  float distance = d/sqrt(2.0);
  float temp[3] = {0};

  temp[planex] = 0; temp[planey] = d;
  displacements[0].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = distance; temp[planey] = distance;
  displacements[1].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = d; temp[planey] = 0;
  displacements[2].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = distance; temp[planey] = -distance;
  displacements[3].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = 0; temp[planey] = -d;
  displacements[4].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = -distance; temp[planey] = -distance;
  displacements[5].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = -d; temp[planey] = 0;
  displacements[6].Set(rtVector(temp[0], temp[1], temp[2]));

  temp[planex] = -distance; temp[planey] = distance;
  displacements[7].Set(rtVector(temp[0], temp[1], temp[2]));
}


void propagateLight(const svConfig& config, int dir, const rtPoint&  light,
		    const float* invol, vector<rtColor>& outvol) {
  const float x = config.volx();
  const float y = config.voly();
  const float z = config.volz();
  const int dims[3] = { x, y, z };
  const float conversion = config.conversion();
  const float scalefactor = config.scaleFactor();
  dlog << "I'm in propagateLight(int)" << endl;

  int i, j, k, index;
  rtPoint currpos, samplept, temppt;
  rtVector currtolight, temptolight, tempvec;
  float distance, directval, indirectval, transparency, cosangle;
  rtColor totalcol(colorLength, false), absfactor(colorLength, false),
          indirectcol(colorLength, false);

  int point[3];
  float temp[3];
  for (/*curr_plane*/; curr_plane < dims[direction] && curr_plane >= 0; curr_plane += viewsign[direction]) {
    
    index = curr_plane*incr[direction];
    
    dlog << "Now propagating to plane " << curr_plane - dims[direction]/2
	 << "\tbase index = " << index << "\tcurr_plane = "
	 << curr_plane << endl;
    
    for (i=0; i<dims[planex]; i++) {
      for (j=0; j<dims[planey]; j++) {
	
	point[direction] = curr_plane-dims[direction]/2;
	point[planex] = i-dims[planex]/2;
	point[planey] = j-dims[planey]/2;
	
	// Current position
	currpos.Set(point[0], point[1], point[2]);
	// Direction to the light, scaled so it falls into the previous plane
	currtolight.Set(currpos, light);
	currtolight.Set(currtolight.Scale(1/fabs(currtolight.getComponent(direction))));
	// Distance to the previous plane
	distance = currtolight.getMagnitude();
	// Position of the center sample
	temp[direction] = curr_plane - viewsign[direction];
	temp[planex] = i + currtolight.getComponent(planex);
	temp[planey] = j + currtolight.getComponent(planey);
	samplept.Set(temp[0], temp[1], temp[2]);
	// Value at the center sample, and light color at that position
	directval = getValues(invol, outvol, /*dims,*/ samplept,
			      &totalcol);
	
	// precalculate the effective distance
	float dist = scalefactor*distance;
	// "transparency" is the proportion of light that doesn't hit the particles
	transparency = exp(-1*directval*conversion*dist);
	// absfactor is the proportion of light that isn't absorbed by the water
	rtColor test(colorLength);
	for (unsigned int channel=0; channel<colorLength; channel++) {
	  absfactor.setChannel(channel,
			       exp(-1*absp[0][colorInd[channel]]*dist));
	}
	
	// Took out albedo, which represented particle scattering
	// Need to put in channel-dependent value for this
	//           totalcol.Set((totalcol*absfactor).Scale(transparency + (1-transparency)*forwardscat));
	totalcol *= absfactor;
	
	totalcol *= transparency + (1-transparency)*innerfactor /* * albedo */;
	
	currtolight.Normalize();
	for (k=0; k<8; k++) {
	  temppt.Set(samplept.getX() + displacements[k].getI(),
		     samplept.getY() + displacements[k].getJ(),
		     samplept.getZ() + displacements[k].getK());
	  if (temppt.getComponent(planex) < 0 ||
	      temppt.getComponent(planey) < 0)
	    continue;
	  
	  indirectval = getValues(invol, outvol, /*dims,*/
				  temppt, &indirectcol);
	  temppt.Set(temppt.getX() - x/2,
		     temppt.getY() - y/2,
		     temppt.getZ() - z/2);
	  distance = currpos.getDistance(temppt);
	  temptolight.Set(temppt, light);
	  temptolight.Normalize();
	  tempvec.Set(currpos, temppt);
	  tempvec.Normalize();
	  cosangle = tempvec.dot(temptolight);
	  
	  // precalculate the effective distance
	  transparency = exp(-1*indirectval*conversion*dist);
	  for (unsigned int channel=0; channel<colorLength; channel++) {
	    absfactor.setChannel(channel,
				 exp(-1*absp[0][colorInd[channel]]*dist));
	  }
	  
	  
	  //             anglescat = phasefactor*pow(1.0 + c*(c - 2.0*cosangle), -1.5);
	  //             indirectcol.Set((indirectcol*absfactor).Scale((1 - transparency)*forwardscat));
	  indirectcol *= absfactor;
	  //             indirectcol *= (1 - transparency)*anglescat*0.125; // Average over 8 samples
	  indirectcol *= (1 - transparency)*outerfactor /* * albedo */ *0.125; // Average over 8 samples
	  
	  totalcol += indirectcol;
	}

	outvol[index].Set(totalcol);
	index += incr[planey];
      }
      index += incr[planex] - dims[planey]*incr[planey];
    }
  }
  
  dlog << "propagateLight(int) completed" << endl;
}



void writeVolume(const string& outFile, unsigned int volume, 
		 const vector<rtColor>& outvol) {

  dlog << "I'm in writeVolume()" << outFile.c_str() << ", ...) " << endl;

  float temp;
  char *marker = reinterpret_cast<char*>(&temp);
#ifdef BIGENDIAN_OUT
  float temp2;
  char *marker2 = reinterpret_cast<char*>(&temp2);
#endif

  ofstream out(outFile.c_str());
  if(!out) {
    dlog << "Could not create file: " << outFile.c_str() << endl;
    exit(-2);
  } else {

    for (unsigned int i=0; i < volume; i++) {
      for (unsigned int channel=0; channel < colorLength; channel++) {
#ifdef BIGENDIAN_OUT
	temp2 = outvol[i].getChannel(channel);
	for (j=0; j<sizeof(float) j++) {
	  marker[j] = marker2[size - j - 1];
	}
#else
	temp = outvol[i].getChannel(channel);
#endif
	out.write(marker, sizeof(float));
	
      }
    }
    out.close();
    dlog << "writeVolume(char*) completed" << endl;
  }
}
