/* Volume Rendering by Ray Casting, using bounding spheres to do space
   leaping */

#include "rtCommon/rtColor.h"
#include "rtCommon/rtPoint.h"
#include "rtCommon/rtVector.h"
#include "rtCommon/rtInterp.h"
#include "rtCommon/rtLightSpec.h"
#include "rtCommon/art.h"
#include "params.h"
#include "renderparams.h"
#include "vrConfig.h"
#include "utilities/readFile.h"
#include "utilities/ddebug.h"
#include <cmath>
#include <vector>
#include <fstream>


void configureLights(const vrConfig&, float*);
rtColor getCumulative(const vrConfig&, rtRay&, float*);
void addBgColor(rtRay&, rtColor&);
void writeImage(const vrConfig&, const vector<rtColor>&);

vector<rtColor> shadow_volume;
int volx,voly,volz;
float f, theta=0, phi=0, psi=0, max_intensity;
rtPoint eye, offset;

float persp = 0;
float oversampling = 1.0;
double visibility = 30;
rtPoint lookat;
bool side=false;

rtColor waterColor(31, false), mineralColor(31, false);
rtColor absorption(colorLength);



#ifdef DEBUG
static DDebug dlog(true, cerr);
#else
static DDebug dlog(false);
#endif



/******** MAIN ***********/

int main(int, char** ) {
  bool exitFlag(false);
  while (!exitFlag) {
    while (!cin) { sleep(60); }
    vrConfig config(cin);
    if (config.error()) {
      dlog << "Error in stdin configuration " << config.error() << endl;
    } else {
      dlog << "Configuration " << config << endl;

      rtColor vox(colorLength, false);
      rtPoint lookat;
      rtPoint origin(0.0, 0.0, 0.0);
      rtVector up(0.0, 0.0, 1.0);
      
      for (int channel=0; channel<31; channel++) {
	waterColor.setChannel(channel, refl[0][channel]);
	mineralColor.setChannel(channel, refl[1][channel]);
      }
      if (colorLength==3) {
	waterColor.Set(waterColor.getRGB());
	mineralColor.Set(mineralColor.getRGB());
      }
      dlog << "Mineral color: " << mineralColor << endl;
      
      volx = config.volx(); voly = config.voly(); volz = config.volz();
      const int dims[3] =  {volx, voly, volz};
      setInterpDims(dims);
      dlog << endl << "Allocating memory..... " << endl;

      float* involume = 0;
      readFile(config.dataFile(), volx*voly*volz, involume);
      
      const rtPoint lightpos = config.lightCenter();
      configureLights(config, involume);
      
      const int xsize = config.imageX(); 
      const int ysize = config.imageY();
      dlog << "image size   x = " << xsize << "  y = " << ysize << endl;
      vector<rtColor> image;
      image.reserve(config.imageX() * config.imageY());
      
      eye.Set(lightpos.getX()+offset.getX(),
	      lightpos.getY()+offset.getY(),
	      lightpos.getZ()+offset.getZ());
      dlog << "Lighting center: " << lightpos << endl;
      dlog << "Offset: " << offset << endl;
      dlog << "Eye position: " << eye << endl;
      if (side)
	lookat.Set(lightpos.getX()+EPSILON-20,
		   lightpos.getY()+EPSILON,
		   lightpos.getZ()+EPSILON);
      else
	lookat.Set(lightpos.getX()+EPSILON,
		   lightpos.getY()+EPSILON+20,
		   lightpos.getZ()+EPSILON);
      dlog << "Lookat position: " << lookat << endl;
      
      rtVector view(eye,lookat);
      view.Normalize();
      if (!view.getMagnitude())
	dlog << "view vector is null\n";
      
      rtVector posx, posy;
      posx.Set(rtVector::cross(view, up));
      posy.Set(rtVector::cross(posx, view));
      
      dlog << "View Vector: " << view << endl
	   << "posX Vector: " << posx << endl
	   << "posY Vector: " << posy << endl;
      dlog << "Now Rendering: " << endl;
      
      rtPoint point;
      rtVector trajectory;
      rtRay ray;
      ray.setColor(rtColor(colorLength));
      ray.setAbs(rtColor(colorLength));
      
      int  xmin, xmax, ymin, ymax;
      xmin = -xsize/2; xmax = (xsize - xsize/2 - 1);
      ymin = -ysize/2; ymax = (ysize - ysize/2 - 1);
      
      // Used for recent images (Jan. 09)
      const float spacing = 0.14 * config.enh(); 
      int count = 1;
      int statusCount = (xsize*ysize > 20000) ? 10000 : 1000;
      float i,j;
      for (j=ymax; j>=ymin; j--) {
	for (i=xmin; i<=xmax; i++) {
	  if (count%statusCount == 0) {
	    dlog << "Casting ray " << count //<< '-' << count + 10000
		 << " out of " << xsize*ysize << endl;
	  }
	  point.Set(eye.getX() + (i*posx.getI() + j*posy.getI())*spacing,
		    eye.getY() + (i*posx.getJ() + j*posy.getJ())*spacing,
		    eye.getZ() + (i*posx.getK() + j*posy.getK())*spacing);
	  
	  // provides perspective view
	  trajectory.setI(view.getI() +
			  ((i/xmax)*posx.getI() + (j/ymax)*posy.getI())*persp);
	  trajectory.setJ(view.getJ() +
			  ((i/xmax)*posx.getJ() + (j/ymax)*posy.getJ())*persp);
	  trajectory.setK(view.getK() +
			  ((i/xmax)*posx.getK() + (j/ymax)*posy.getK())*persp);
	  trajectory.Normalize();
	  trajectory.Scale(1.0/oversampling);
	  
	  ray.SetLocation(point,trajectory);
	  ray.resetColors();
	  count++;
	  vox = getCumulative(config, ray, involume);
	  image.push_back(vox.getRGB());
	}
      }
      
      writeImage(config, image);

      if (involume) delete [] involume;
      exitFlag = config.exit();
      // Specifically on cout for chaining.
      cout << config << endl << (config.exit() ? "exit" : "done") << endl;
      
    } // end of processing for good config
  } // while 1.  will only get here if the exit flag is set

  return 0;
}
    


  /*************************
   ******* Functions *******
  *************************/

void configureLights(const vrConfig& config,  float* data) {

    /******************************/
  /*   READ IN THE LIGHT DATA   */
  /******************************/
  // Assuming that the light volumes are the same size and color length
  // as the data volume, read in the light data
  const int size = config.volx() * config.voly() * config.volz();
  shadow_volume.reserve(size);

  int k, x=0, y=0, z=0;
  float rgb[colorLength]
      , /*eyeprod,*/ lightprod;
  rtColor readcolor(colorLength,false), totalcolor(colorLength,false);
  rtColor directcolor(colorLength,false), directtotal(colorLength,false);
  rtVector normal, lightdir, eyedir;

  // Find out how many light volumes there will be
  const vector<rtLightSpec> lights = config.lights();
  const int n = lights.size();
  vector<ifstream*> shadowvols;
  vector<rtPoint> positions;

  // Open all the light volume files, and record the light positions
  vector<rtLightSpec>::const_iterator i;
  for (vector<rtLightSpec>::const_iterator i = lights.begin(); 
       i < lights.end(); i++) {
    ifstream* tempstream = new ifstream(i->file().c_str());
    if (!tempstream) {
      dlog << "Unable to open file: " << i->file().c_str() << endl;
      exit(1);
    }
    shadowvols.push_back(tempstream);

    rtPoint temp(i->position().getX() + volx/2.0,
		 i->position().getY() + voly/2.0,
		 i->position().getZ() + volz/2.0);
    positions.push_back(temp);
  }

  rtColor minCol(colorLength, false), maxCol(colorLength, false);
  // Since we know the eye and light positions,
  // precalculate and combine the light contributions at each voxel
  dlog << "starting loop for combining light contributions " << endl;
  for( int i=0; i<size; i++) {
    totalcolor.setValues(0.0);
    directtotal.setValues(0.0);
    rtVector normal(getNormal(data, i));
    normal.Normalize();
    eyedir.setI(eye.getX() - x);
    eyedir.setJ(eye.getY() - y);
    eyedir.setK(eye.getZ() - z);
    eyedir.Normalize();
    if (!eyedir.getMagnitude()) {
      dlog << "eyedir vector is null at " << rtPoint(x,y,z) << endl;
    }

    // Get the contribution from each light source
    for (k=0; k<n; k++) {
      // get the direction TO the light
      lightdir.setI(positions[k].getX() - x);
      lightdir.setJ(positions[k].getY() - y);
      lightdir.setK(positions[k].getZ() - z);
      lightdir.Normalize();
      if (!lightdir.getMagnitude()) {
        dlog << "lightdir is null for light " << k << " at " 
	     << rtPoint(x,y,z) << endl;
      }

      // diffuse phong shading
      float LdotN = -lightdir.dot(normal);
      LdotN = (LdotN < 0) ? 0 : LdotN;
      lightprod = config.ambient() + config.diffuse() * LdotN;

      shadowvols[k]->read((char*) rgb, colorLength*sizeof(float));

      for (unsigned int channel=0; channel<colorLength; channel++) {
#ifdef BIGENDIAN_LIGHT
	char* marker = (char*) rgb;
        for (j=0;j<sizeof(float);j++) {
          temp[j] = marker[sizeof(float) - j - 1];
        }
        rgb[channel] = *((float*) temp);
#endif
        if (isnan(rgb[channel]) || isinf(rgb[channel])) {
          dlog << "Error reading light file " << k << " at index " << i << endl;
          exit(-1);
        }
        if (rgb[channel] < minCol.getChannel(channel))
          minCol.setChannel(channel, rgb[channel]);
        if (rgb[channel] > maxCol.getChannel(channel))
          maxCol.setChannel(channel, rgb[channel]);
        readcolor.setChannel(channel, rgb[channel]);
      }

      float directFactor = eyedir.dot(lightdir);
      if (directFactor < 0) {
        directFactor = -directFactor;
      } else {
        directFactor = 0;
      }

      directcolor = readcolor;
      directcolor *= directFactor;
      directtotal += directcolor;

	    // Accumulate the light contribution
      readcolor *= lightprod;
      totalcolor += readcolor;
    }

    shadow_volume.push_back(totalcolor);
    if (x == volx-1) {
      x=0;
      if (y == voly-1) {
        y=0;
        z++;
      } else
        y++;
    }
    else
      x++;
  }

  for (vector<ifstream*>::iterator i= shadowvols.begin(); 
       i<shadowvols.end(); i++) {
    (*i)->close();
    delete *i;
  }
}




rtColor getCumulative(const vrConfig& config, rtRay& ray, float* data) {
  const float colorscale = config.colorscale();
  const float scalefactor = config.scaleFactor();
  const float reflFactorM = config.reflFactorM();
  const float reflFactorW = config.reflFactorW();
  const float conversion = config.conversion();

  // for direction, 1 indicates non-negative and -1 indicates negative
  char xdir=1, ydir=1, zdir=1;
  int i=0;
  rtColor color(colorLength, false), water(colorLength, false), light(colorLength, false);
  double density, /*visfactor=0,*/ sampleDist;
  float alpha, alpha_min/*, angle, reflectivity=1*/;
  rtPoint pos;
  rtVector lightdir, normal;
  float test = -1;

  if (ray.getI() < 0)
    xdir = -1;
  if (ray.getJ() < 0)
    ydir = -1;
  if (ray.getK() < 0)
    zdir = -1;

  pos.Set(ray.getX() + ray.getI() + volx/2.0,
          ray.getY() + ray.getJ() + voly/2.0,
          ray.getZ() + ray.getK() + volz/2.0);
  sampleDist = ray.getDirection().getMagnitude() * scalefactor;
  alpha_min = 1 - exp(-1 * density_min * conversion * sampleDist);
  for (unsigned int channel=0; channel<colorLength; channel++) {
    absorption.setChannel(channel,
                          exp(-1*absp[0][colorInd[channel]]*sampleDist));
  }
  float waterAlpha = 1 - exp(-1*sampleDist*reflFactorW);

// Figure out how many times the ray must be incremented
// before all three coordinates have entered the volume

  float sampleDir[3] = {ray.getI(),ray.getJ(),ray.getK()};
  int dirs[3] = {xdir,ydir,zdir};
  int dims[3] = {volx,voly,volz};
  int tests[3] = {0};
  float coords[3] = {pos.getX(),pos.getY(),pos.getZ()};
  test=0;
  for (i=0; i<3 && test > -1; i++) {
    if (coords[i]<0) {
      if (dirs[i]>0) {
        tests[i] = 1 + (-coords[i]/sampleDir[i]);
      }
      else
        test = -1;
    }
    else if (coords[i]>(dims[i]-1)) {
      if (dirs[i]<0) {
        tests[i] = 1 + (-(coords[i] - (dims[i]-1))/sampleDir[i]);
      }
      else
        test = -1;
    }
  }


  if (test == -1) {
    // supressed this message, 'cause there are a lot of them.
    // dlog << "The ray points away from the data.\n";
    color.setValues(0.0);
    return color;
  }
  else  {
    for (i=0; i<3; i++) {
      if (tests[i] > test)
        test = tests[i];
    }
    if (test) {
      float xx, yy, zz;
      xx = pos.getX() + test*ray.getI();
      yy = pos.getY() + test*ray.getJ();
      zz = pos.getZ() + test*ray.getK();
      if ((xx>volx-1)||(yy>voly-1)||(zz>volz-1)||(xx<0)||(yy<0)||(zz<0)) {
      // There is no significant intersection
        color.setValues(0.0);
        return color;
      }
    }
  }

  /* Process empty water until the volume is reached */
  for (i=0; i<test; i++) {
	  // process_BGpoint(ray);
    if (ray.getAlpha() > (1 - EPSILON)) {
      dlog << "The ray reached opacity.\n";
      return ray.getColor().Scale(colorscale);
    }

    ray.multAbs(absorption);
    alpha = alpha_min * (1.0 - ray.getAlpha());
    ray.addAlpha(alpha);

    pos.setX(pos.getX() + ray.getI());
    pos.setY(pos.getY() + ray.getJ());
    pos.setZ(pos.getZ() + ray.getK());
  }

  // Traverse the volume
  int samplecount = 0;
  rtColor final(colorLength);
  while((pos.getX()<volx)&&(pos.getY()<voly)&&(pos.getZ()<volz) &&
        (pos.getX()>=0)&&(pos.getY()>=0)&&(pos.getZ()>=0))
  {

    /* get all the parameters and find the new color! */
    density = getValues(data, shadow_volume, /*dims,*/ pos, &light);
    alpha = 1 - exp(-1 * density * conversion * sampleDist); // opacity

    //////
    // Determine the local reflected color
    //////

    color.Set(mineralColor);
    color *= alpha*reflFactorM; // proportion of mineral that reflects

    water.Set(waterColor);
    water *= (1.0 - alpha)*waterAlpha; // proportion of water that reflects
    alpha += (1.0 - alpha)*waterAlpha; // water reflection contributes to opacity

    color += water; // total reflective color at the sample point

    //////
    // Calculated the reflected color that reaches the eye
    //////

    color *= light; // total reflected color at the sample point
//     color = light; color *= 0.1;
    color *= ray.getAbs(); // the color of reflected light that can reach the eye without being absorbed
    color *= 1.0 - ray.getAlpha(); // the intensity of reflected color that can reach the eye without being occluded


    // Add the calculated contribution
    ray.addColor(color);

    //////
    // Accumulate the total alpha and absorption
    //////

    ray.multAbs(absorption); // Accumulate the absorption
    alpha *= 1.0 - ray.getAlpha(); // alpha reduces only light that made it this far
    ray.addAlpha(alpha); // Accumulate alpha

    // Check whether the ray reached opacity
    if (ray.getAlpha() > (1 - EPSILON)) {
      return ray.getColor().Scale(colorscale);
    }

    pos.setX(pos.getX() + ray.getI());
    pos.setY(pos.getY() + ray.getJ());
    pos.setZ(pos.getZ() + ray.getK());
    i++;
    samplecount++;
  }

  addBgColor(ray, light);

  final.Set(ray.getColor());
  final *= colorscale;
  return final;
}



void addBgColor(rtRay& ray, rtColor& light) {
  rtColor myBg(colorLength);
  for (unsigned int channel=0; channel<colorLength; channel++)
    myBg.setChannel(channel, bg[colorInd[channel]]);

  myBg *= light; // total reflected color at the sample point
  myBg *= ray.getAbs(); // the color of reflected light that can reach the eye without being absorbed
  myBg *= 1.0 - ray.getAlpha(); // the intensity of reflected color that can reach the eye without being occluded

  ray.addColor(myBg);
}


void writeImage(const vrConfig& config, const vector<rtColor>& image ) {
  ofstream file(config.imageFile().c_str());
  if(!file) {
    dlog << "Could not create file: " << config.imageFile().c_str();
  } else {
    dlog << "writting file " << config.imageFile().c_str() << endl;
    int size = config.imageX() * config.imageY();
    dlog << "# of pixels: " << size << "    image.size(): "
	 << image.size() << endl;
    
    file << "P6" << endl << config.imageY() << " " << config.imageX()
	 << endl << "255" << endl;

    for(unsigned int w=0; w<image.size();w++) {
      file.put(image[w].putR());
      file.put(image[w].putG());
      file.put(image[w].putB());
    }
    file.close();
  }
}
  
